/* ppmtoicr.c - convert a portable pixmap to NCSA ICR protocol
**
** Copyright (C) 1990 by Kanthan Pillay (svpillay@Princeton.EDU)
** Portions Copyright (C) 1989 by Jef Poskanzer.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/

#include <stdio.h>
#include "ppm.h"
#include "ppmcmap.h"

#define MAXCOLORS 256
#define CLUTCOLORS 768

pixel **pixels;
colorhash_table cht;
char *malloc(), *testimage;

main(argc, argv)
int argc;
char *argv[];
{
	FILE *ifd;
	int argn, rows, cols, colors, i, j, BitsPerPixel, newxsize;
	pixval maxval;
	colorhist_vector chv;
	int Red[MAXCOLORS], Green[MAXCOLORS], Blue[MAXCOLORS];
	char rgb[CLUTCOLORS];
	int GetPixel();
	char *usage = "[-r] [-w windowname] [-d display] [-e expand] [ppmfile]";
	char *windowname, *thischar, *thisline, *space;
	register unsigned char c;
	register char *p;
	int ci, icolors;
	int display = 0, expand = 1;

	int rflag = 0, wflag = 0, errflg = 0;
	int q, z;
	extern int optind, opterr;
	extern char *optarg;

/*	I use getopt(3) for parsing options. If you don't have it, there's a
	public domain version available from comp.sources.misc  */

	pm_progname = argv[0];
	windowname = "untitled";
	opterr = 0;
	while ((q = getopt(argc, argv, "rw:d:e:")) != EOF)
	switch (q) {
	case 'r':
		rflag++;
	break;
	case 'w':
		windowname=optarg;
		wflag++;
	break;
	case 'd':
		z = (sscanf(optarg,"%d",&display));
		if ((z == EOF) || (z == 0))
			errflg++;
	break;
	case 'e':
		z = (sscanf(optarg,"%d",&expand));
		if ((z == EOF) || (z == 0))
			errflg++;
	break;
	case '?':
		errflg++;
	}

	if (errflg) {
		pm_usage(usage);
		exit(2);
	}

	if (optind < argc)
	{
	ifd = pm_openr(argv[optind]);
	if (!wflag)
		windowname = argv[optind];
	optind++;
	}
	else
	{
	ifd = stdin;
	}

	if (optind != argc)
	pm_usage(usage);

	pixels = ppm_readppm(ifd, &cols, &rows, &maxval);

	pm_close(ifd);

	for (i = 0; i < CLUTCOLORS; i++)
	rgb[i] = 0;

	/* Figure out the colormap. This is Jeff's code from ppmtogif.c */
	fprintf(stderr, "(Computing colormap..." );
	fflush(stderr);
	chv = ppm_computecolorhist(pixels, cols, rows, MAXCOLORS, &colors);
	if (chv == (colorhist_vector) 0)
	pm_error(
		"too many colors - try running the pixmap through 'ppmquant 256'",
		0,0,0,0,0 );
	fprintf(stderr, "Done. %d colors found.)\n", colors);

	/* Turn the ppm colormap into an ICR colormap. */
	if (maxval > 255)
	fprintf(
		stderr, "(Maxval is not 255 -- automatically rescaling colors.)\n");
	for (i = 0; i < colors; i++)
	{
	j = (3 * i);
	if (maxval == 255)
		{
		rgb[j] = PPM_GETR(chv[i].color) ;
		j++;
		rgb[j] = PPM_GETG(chv[i].color) ;
		j++;
		rgb[j] = PPM_GETB(chv[i].color) ;
		}
	else
		{
		rgb[j] = (int) PPM_GETR(chv[i].color) * 255 / maxval;
		j++;
		rgb[j] = (int) PPM_GETG(chv[i].color) * 255 / maxval;
		j++;
		rgb[j] = (int) PPM_GETB(chv[i].color) * 255 / maxval;
		}
	}
	BitsPerPixel = colorstobpp(colors);

	/* And make a hash table for fast lookup. Jeff's code again. */
	cht = ppm_colorhisttocolorhash(chv, colors);
	ppm_freecolorhist(chv);


	/************** Create a new window using ICR protocol *********/
	/* Format is "ESC^W;left;top;width;height;display;windowname"  */

	fprintf(stderr, "(Creating window %s ...",windowname);
	(void)printf("\033^W;%d;%d;%d;%d;%d;%s^",0,0,cols*expand,rows*expand,display,windowname);
	fflush(stdout);
	fprintf(stderr, "Done ...)\n");


	/****************** Download the colormap.  ********************/
	fprintf(stderr, "(Downloading colormap for %s ...",windowname);

	(void)printf("\033^M;%d;%d;%d;%s^",0,MAXCOLORS,CLUTCOLORS,windowname);
	thischar = rgb;
	for (j=0; j<CLUTCOLORS; j++) {
	c = *thischar++;
		if (c > 31 && c < 123 ) {	 /* printable ASCII */
		putchar(c);
		}
		else {
		putchar((c>>6)+123);	 /* non-printable, so encode it */
		putchar((c & 0x3f) + 32);
		}
	}
	fflush(stdout);
	fprintf(stderr, "Done ...)\n");

	/**************** send out picture *************************/
	/* Protocol's RLE scheme is quicker but buggy              */

	if (rflag) {	
		fprintf(stderr, "(Sending run-length encoded picture data ...");
		testimage = malloc(rows*cols);
		p = testimage;
		for (i=0; i<rows; i++)
			for (j=0; j<cols; j++) 
			*p++ = GetPixel(j,i);
		space = malloc(rows*3);
		thisline = testimage;
		for (i = 0; i < rows; i++) {
			newxsize = rleit(thisline,space,cols);
			thisline += cols;	/* increment to next line */
		(void)printf("\033^R;%d;%d;%d;%d;%s^",0,i*expand,expand,newxsize,windowname);
		thischar = space;
		for (j=0; j< newxsize; j++) {
			c= *thischar++;  /*get byte to send */
			if (c>31 && c <123) {
				putchar(c);
				}
			else {
				putchar((c>>6) + 123);
				putchar((c & 0x3f) + 32);
				}
			}
			fflush(stdout);
		}
		free(space);
		fprintf(stderr, "Done ...)\n");
		exit(0);
		}

	/* Otherwise, send out uncompressed pixel data via the slow method */

		else {
		fprintf(stderr, "(Sending picture data ...");
		for (i = 0; i < rows; i++) {
			(void)printf("\033^P;%d;%d;%d;%d;%s^",0,i*expand,expand,cols,windowname);
			for (j = 0; j < cols; j++) {
				c  = GetPixel(j,i);
				if (c > 31 && c < 123) {
						putchar(c);
						}
				else		{
						putchar((c>>6)+123);
						putchar((c & 0x3f) + 32);
						}
				}
			}
		fflush(stdout);
		fprintf(stderr, "Done)\n");
		exit(0);
		}
	}

/* Jeff's code */
colorstobpp(colors)
int colors;
	{
	int bpp;

	if (colors <= 2)
	bpp = 1;
	else if (colors <= 4)
	bpp = 2;
	else if (colors <= 8)
	bpp = 3;
	else if (colors <= 16)
	bpp = 4;
	else if (colors <= 32)
	bpp = 5;
	else if (colors <= 64)
	bpp = 6;
	else if (colors <= 128)
	bpp = 7;
	else if (colors <= 256)
	bpp = 8;
	else
	pm_error("can't happen", 0,0,0,0,0);
	return bpp;
	}

GetPixel(x, y)
int x, y;
	{
	int color;

	color = ppm_lookupcolor(cht, pixels[y][x]);
	return color;
	}



/* rleit   compress with run length encoding as per NCSA's documentation */

rleit(buf,bufto,len)
	int len;
	char *buf,*bufto;
	{
	register char *p,*q,*cfoll,*clead;
	char *begp;
	int i;

	p = buf;
	cfoll = bufto;
	clead = cfoll + 1;

	begp = p;
	while (len > 0 ) {		/* encode until gone */
		
		q = p + 1;
		i = len-1;
	while (*p == *q && i+120 > len && i) {
		q++;
		i--;
	}

	if (q > p +2) {			/* three in a row */
		if (p > begp) {
			*cfoll = p - begp;
			cfoll = clead;
		}
		*cfoll++ = 128 | (q-p);		/*len of seq*/
		*cfoll++ = *p;			/* char of seq */
		len -= q-p;		/* subtract len of seq */
		p = q;
		clead = cfoll+1;
		begp = p;
	}
	else {
		*clead++ = *p++;	/* copy one char */
		len--;
		if (p>begp + 120) {
			*cfoll = p - begp;
			cfoll = clead++;
			begp = p;
		}
	}
	}

/* fillin last bytecount */

	if (p>begp)
		*cfoll = 128 | (p-begp);
	else
		clead--;

	return((int) (clead-bufto));	/*how many stored as encoded */
}


/* End */
